# Command line entry point for candump2cancat
# This requires scapy to be installed

import sys
import argparse
from cancat.utils import convert


def main():
    argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
            prog='pcap2cancat',
            description='Utility to convert a pcap with CAN messages into a CanCat session')
    parser.add_argument('pcap', help='input pcap')
    parser.add_argument('output', help='output cancat session')
    args = parser.parse_args(argv)

    convert.pcap2cancat(args.pcap, args.output)



# Utility functions for CanCat

import pickle
import cancat
import struct
import re


def cancat2candump(session, output):
    with open(session, 'rb') as f:
        sess = pickle.load(f)

    with open(output, 'w') as f:
        for msg_time, msg in sess['messages'].get(cancat.CMD_CAN_RECV, None):
            line = '({:.6f}) vcan0 {}#{}\n'.format(
                msg_time,
                msg[0:4].encode('HEX'),
                msg[4:].encode('HEX')
            )
            f.write(line)


def cancat2pcap(session, output):
    import scapy.layers.l2
    import scapy.packet
    import scapy.utils

    sess = pickle.load(session)

    msgs = []
    for msg_time, msg in sess['messages'].get(cancat.CMD_CAN_RECV, None):
        arb = struct.unpack_from('>L', msg)[0] | 0x80000000
        msg_data = msg[4:]
        msg_len = len(msg_data)
        raw = struct.pack('<LL', arb, msg_len) + msg_data + '\x00' * (8 - msg_len)

        pkt = scapy.layers.l2.CookedLinux(
                pkttype=1,
                lladdrtype=0x118,
                lladdrlen=0,
                src='',
                proto=0xc
              ) / scapy.packet.Raw(
                load=raw
              )
        pkt.time = msg_time

        msgs.append(pkt)

    scapy.utils.wrpcap(output, msgs)


def _import_candump(filename):
    msgs = []
    with open(filename, 'r') as f:
        pat = re.compile(r'\(([0-9]+\.[0-9]+)\) [A-Za-z0-9]+ ([A-Fa-f0-9]+)#([A-Fa-f0-9]+)')
        for line in f.readlines():
            match = pat.match(line.strip())
            if match is None:
                raise ValueError('Invalid candump format: %s' % line)

            time, arb_id, data = match.groups()

            # Ensure that the arbid is padded out to 4 bytes
            if len(arb_id) < 8:
                arb_id = ('0' * (8 - len(arb_id))) + arb_id

            msgs.append((float(time), arb_id.decode('hex') + data.decode('hex')))

    sess = {
        'bookmark_info': {},
        'bookmarks': [],
        'comments': [],
        'messages': {
            cancat.CMD_CAN_RECV: msgs,
        },
    }

    return sess


def _import_pcap(filename):
    import scapy.layers.l2
    import scapy.packet
    import scapy.utils

    msgs = []
    can_pkts = [p for p in scapy.utils.rdpcap(filename) \
            if scapy.layers.l2.CookedLinux in p and p.proto == 12]
    for pkt in can_pkts:
        [arb_id, data_len] = struct.unpack_from('<LL', pkt.load)

        # Clear any flags in the arbitration ID field that might be set
        arb_id &= 0x1FFFFFFF
        msgs.append((pkt.time, struct.pack('>L', arb_id) + pkt.load[8:8+data_len]))

    sess = {
        'bookmark_info': {},
        'bookmarks': [],
        'comments': [],
        'messages': {
            cancat.CMD_CAN_RECV: msgs,
        },
    }

    return sess


def candump2cancat(candumplog, output):
    with open(output, 'w') as f:
        pickle.dump(_import_candump(candumplog), f)


def pcap2cancat(pcapfile, output):
    with open(output, 'w') as f:
        pickle.dump(_import_pcap(pcapfile), f)



import vstruct

from vstruct.primitives import *
from vstruct.defs.macho.const import *

vm_prot_t = v_uint32
cpu_type_t = v_uint32
cpu_subtype_t = v_uint32
lc_str = v_uint32

class mach_header(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.magic      = v_uint32() # mach magic number identifier
        self.cputype    = cpu_type_t() # cpu specifier
        self.cpusubtype = cpu_subtype_t() # machine specifier
        self.filetype   = v_uint32() # type of file
        self.ncmds      = v_uint32() # number of load commands
        self.sizeofcmds = v_uint32() # the size of all the load commands
        self.flags      = v_uint32() # flags

    def vsParse(self, bytes, offset=0):
        # Over-ride this so we can do the parse, and make sure we
        # had the right endianness.
        ret = vstruct.VStruct.vsParse(self, bytes, offset=offset)
        if self.magic == MH_CIGAM:
            self._vs_fmtbase = '>'
            ret = vstruct.VStruct.vsParse(self, bytes, offset=offset)
        return ret

class mach_header_64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.magic      = v_uint32() # mach magic number identifier
        self.cputype    = cpu_type_t() # cpu specifier
        self.cpusubtype = cpu_subtype_t() # machine specifier
        self.filetype   = v_uint32() # type of file
        self.ncmds      = v_uint32() # number of load commands
        self.sizeofcmds = v_uint32() # the size of all the load commands
        self.flags      = v_uint32() # flags
        self.reserved   = v_uint32() # reserved


# FIXME all commands should subclass this one!
class load_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # type of load command
        self.cmdsize = v_uint32() # total size of command in bytes

class segment_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd         = v_uint32() # LC_SEGMENT
        self.cmdsize     = v_uint32() # includes sizeof section structs
        self.segname     = v_str(size=16) # segment name
        self.vmaddr      = v_uint32() # memory address of this segment
        self.vmsize      = v_uint32() # memory size of this segment
        self.fileoff     = v_uint32() # file offset of this segment
        self.filesize    = v_uint32() # amount to map from the file
        self.maxprot     = vm_prot_t() # maximum VM protection
        self.initprot    = vm_prot_t() # initial VM protection
        self.nsects      = v_uint32() # number of sections in segment
        self.flags       = v_uint32() # flags

class segment_command_64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd         = v_uint32() # LC_SEGMENT_64
        self.cmdsize     = v_uint32() # includes sizeof section_64 structs
        self.segname[16] = v_uint8() # segment name
        self.vmaddr      = v_uint64() # memory address of this segment
        self.vmsize      = v_uint64() # memory size of this segment
        self.fileoff     = v_uint64() # file offset of this segment
        self.filesize    = v_uint64() # amount to map from the file
        self.maxprot     = vm_prot_t() # maximum VM protection
        self.initprot    = vm_prot_t() # initial VM protection
        self.nsects      = v_uint32() # number of sections in segment
        self.flags       = v_uint32() # flags


class section(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.sectname     = v_str(size=16) # name of this section
        self.segname      = v_str(size=16) # segment this section goes in
        self.addr         = v_uint32() # memory address of this section
        self.size         = v_uint32() # size in bytes of this section
        self.offset       = v_uint32() # file offset of this section
        self.align        = v_uint32() # section alignment (power of 2)
        self.reloff       = v_uint32() # file offset of relocation entries
        self.nreloc       = v_uint32() # number of relocation entries
        self.flags        = v_uint32() # flags (section type and attributes)
        self.reserved1    = v_uint32() # reserved (for offset or index)
        self.reserved2    = v_uint32() # reserved (for count or sizeof)


class section_64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.sectname     = v_str(size=16) # name of this section
        self.segname      = v_str(size=16) # segment this section goes in
        self.addr         = v_uint64() # memory address of this section
        self.size         = v_uint64() # size in bytes of this section
        self.offset       = v_uint32() # file offset of this section
        self.align        = v_uint32() # section alignment (power of 2)
        self.reloff       = v_uint32() # file offset of relocation entries
        self.nreloc       = v_uint32() # number of relocation entries
        self.flags        = v_uint32() # flags (section type and attributes)
        self.reserved1    = v_uint32() # reserved (for offset or index)
        self.reserved2    = v_uint32() # reserved (for count or sizeof)
        self.reserved3    = v_uint32() # reserved

class fvmlib_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_IDFVMLIB or LC_LOADFVMLIB
        self.cmdsize = v_uint32() # includes pathname string
        self.name          = lc_str() # library's target pathname
        self.minor_version = v_uint32() # library's minor version number
        self.header_addr   = v_uint32() # library's header address

class dylib_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_ID_DYLIB, LC_LOAD_{,WEAK_}DYLIB, LC_REEXPORT_DYLIB
        self.cmdsize = v_uint32() # includes pathname string
        self.name                  = lc_str() # library's path name
        self.timestamp             = v_uint32() # library's build time stamp
        self.current_version       = v_uint32() # library's current version number
        self.compatibility_version = v_uint32() # library's compatibility vers number
        self.namedata              = v_bytes(size=0)

    def vsParse(self, bytes, offset=0):
        # So we can grab the name data
        retoff = vstruct.VStruct.vsParse(self, bytes, offset=offset)
        # Grab the name from the inline data...
        name = bytes[ offset + self.name : offset + self.cmdsize ]
        name = name.split('\x00', 1)[0]
        self.vsGetField('namedata').vsSetLength(len(name))
        self.namedata = name
        return retoff

class sub_framework_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd      = v_uint32() # LC_SUB_FRAMEWORK
        self.cmdsize  = v_uint32() # includes umbrella string
        self.umbrella = lc_str() # the umbrella framework name


class sub_client_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_SUB_CLIENT
        self.cmdsize = v_uint32() # includes client string
        self.client  = lc_str() # the client name


class sub_umbrella_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd          = v_uint32() # LC_SUB_UMBRELLA
        self.cmdsize      = v_uint32() # includes sub_umbrella string
        self.sub_umbrella = lc_str() # the sub_umbrella framework name


class sub_library_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd         = v_uint32() # LC_SUB_LIBRARY
        self.cmdsize     = v_uint32() # includes sub_library string
        self.sub_library = lc_str() # the sub_library name


class prebound_dylib_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd            = v_uint32() # LC_PREBOUND_DYLIB
        self.cmdsize        = v_uint32() # includes strings
        self.name           = lc_str() # library's path name
        self.nmodules       = v_uint32() # number of modules in library
        self.linked_modules = lc_str() # bit vector of linked modules


class dylinker_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_ID_DYLINKER or LC_LOAD_DYLINKER
        self.cmdsize = v_uint32() # includes pathname string
        self.name    = lc_str() # dynamic linker's path name


class thread_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_THREAD or LC_UNIXTHREAD
        self.cmdsize = v_uint32() # total size of this command


class routines_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd          = v_uint32() # LC_ROUTINES
        self.cmdsize      = v_uint32() # total size of this command
        self.init_address = v_uint32() # address of initialization routine
        self.init_module  = v_uint32() # index into the module table that
        self.reserved1    = v_uint32()
        self.reserved2    = v_uint32()
        self.reserved3    = v_uint32()
        self.reserved4    = v_uint32()
        self.reserved5    = v_uint32()
        self.reserved6    = v_uint32()


class routines_command_64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd          = v_uint32() # LC_ROUTINES_64
        self.cmdsize      = v_uint32() # total size of this command
        self.init_address = v_uint64() # address of initialization routine
        self.init_module  = v_uint64() # index into the module table that
        self.reserved1    = v_uint64()
        self.reserved2    = v_uint64()
        self.reserved3    = v_uint64()
        self.reserved4    = v_uint64()
        self.reserved5    = v_uint64()
        self.reserved6    = v_uint64()


class symtab_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_SYMTAB
        self.cmdsize = v_uint32() # sizeof(struct symtab_command)
        self.symoff  = v_uint32() # symbol table offset
        self.nsyms   = v_uint32() # number of symbol table entries
        self.stroff  = v_uint32() # string table offset
        self.strsize = v_uint32() # string table size in bytes

class dysymtab_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd            = v_uint32() # LC_DYSYMTAB
        self.cmdsize        = v_uint32() # sizeof(struct dysymtab_command)
        self.ilocalsym      = v_uint32() # index to local symbols
        self.nlocalsym      = v_uint32() # number of local symbols
        self.iextdefsym     = v_uint32() # index to externally defined symbols
        self.nextdefsym     = v_uint32() # number of externally defined symbols
        self.iundefsym      = v_uint32() # index to undefined symbols
        self.nundefsym      = v_uint32() # number of undefined symbols
        self.tocoff         = v_uint32() # file offset to table of contents
        self.ntoc           = v_uint32() # number of entries in table of contents
        self.modtaboff      = v_uint32() # file offset to module table
        self.nmodtab        = v_uint32() # number of module table entries
        self.extrefsymoff   = v_uint32() # offset to referenced symbol table
        self.nextrefsyms    = v_uint32() # number of referenced symbol table entries
        self.indirectsymoff = v_uint32() # file offset to the indirect symbol table
        self.nindirectsyms  = v_uint32() # number of indirect symbol table entries
        self.extreloff      = v_uint32() # offset to external relocation entries
        self.nextrel        = v_uint32() # number of external relocation entries
        self.locreloff      = v_uint32() # offset to local relocation entries
        self.nlocrel        = v_uint32() # number of local relocation entries


class dylib_table_of_contents(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.symbol_index = v_uint32() # the defined external symbol (index into the symbol table)
        self.module_index = v_uint32() # index into the module table this symbol is defined in


class dylib_module(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.module_name           = v_uint32() # the module name (index into string table)
        self.iextdefsym            = v_uint32() # index into externally defined symbols
        self.nextdefsym            = v_uint32() # number of externally defined symbols
        self.irefsym               = v_uint32() # index into reference symbol table
        self.nrefsym               = v_uint32() # number of reference symbol table entries
        self.ilocalsym             = v_uint32() # index into symbols for local symbols
        self.nlocalsym             = v_uint32() # number of local symbols
        self.iextrel               = v_uint32() # index into external relocation entries
        self.nextrel               = v_uint32() # number of external relocation entries
        self.iinit_iterm           = v_uint32() # low 16 bits are the index into the init section, high 16 bits are the index into the term section
        self.ninit_nterm           = v_uint32() # low 16 bits are the number of init section entries, high 16 bits are the number of term section entries
        self.objc_module_info_addr = v_uint32() # the (__OBJC,__module_info) section
        self.objc_module_info_size = v_uint32() # the (__OBJC,__module_info) section


class dylib_module_64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.module_name           = v_uint32() # the module name (index into string table)
        self.iextdefsym            = v_uint32() # index into externally defined symbols
        self.nextdefsym            = v_uint32() # number of externally defined symbols
        self.irefsym               = v_uint32() # index into reference symbol table
        self.nrefsym               = v_uint32() # number of reference symbol table entries
        self.ilocalsym             = v_uint32() # index into symbols for local symbols
        self.nlocalsym             = v_uint32() # number of local symbols
        self.iextrel               = v_uint32() # index into external relocation entries
        self.nextrel               = v_uint32() # number of external relocation entries
        self.iinit_iterm           = v_uint32() # low 16 bits are the index into the init section, high 16 bits are the index into the term section
        self.ninit_nterm           = v_uint32() # low 16 bits are the number of init section entries, high 16 bits are the number of term section entries
        self.objc_module_info_size = v_uint32() # the (__OBJC,__module_info) section
        self.objc_module_info_addr = v_uint64() # the (__OBJC,__module_info) section


class dylib_reference(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.flags = v_uint32() # flags to indicate the type of reference


class twolevel_hints_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_TWOLEVEL_HINTS
        self.cmdsize = v_uint32() # sizeof(struct twolevel_hints_command)
        self.offset  = v_uint32() # offset to the hint table
        self.nhints  = v_uint32() # number of hints in the hint table


class twolevel_hint(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.itoc = v_uint32() # index into the table of contents

class prebind_cksum_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_PREBIND_CKSUM
        self.cmdsize = v_uint32() # sizeof(struct prebind_cksum_command)
        self.cksum   = v_uint32() # the check sum or zero


class uuid_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd      = v_uint32() # LC_UUID
        self.cmdsize  = v_uint32() # sizeof(struct uuid_command)
        self.uuid[16] = v_uint8() # the 128-bit uuid


class rpath_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_RPATH
        self.cmdsize = v_uint32() # includes string
        self.path    = lc_str() # path to add to run path


class linkedit_data_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd      = v_uint32() # LC_CODE_SIGNATURE or LC_SEGMENT_SPLIT_INFO
        self.cmdsize  = v_uint32() # sizeof(struct linkedit_data_command)
        self.dataoff  = v_uint32() # file offset of data in __LINKEDIT segment
        self.datasize = v_uint32() # file size of data in __LINKEDIT segment


class encryption_info_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd       = v_uint32() # LC_ENCRYPTION_INFO
        self.cmdsize   = v_uint32() # sizeof(struct encryption_info_command)
        self.cryptoff  = v_uint32() # file offset of encrypted range
        self.cryptsize = v_uint32() # file size of encrypted range
        self.cryptid   = v_uint32() # which enryption system, 0 means not-encrypted yet


class symseg_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_SYMSEG
        self.cmdsize = v_uint32() # sizeof(struct symseg_command)
        self.offset  = v_uint32() # symbol segment offset
        self.size    = v_uint32() # symbol segment size in bytes


class ident_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd     = v_uint32() # LC_IDENT
        self.cmdsize = v_uint32() # strings that follow this command

class fvmfile_command(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.cmd         = v_uint32() # LC_FVMFILE
        self.cmdsize     = v_uint32() # includes pathname string
        self.name        = lc_str() # files pathname
        self.header_addr = v_uint32() # files virtual address

command_classes = {
    LC_SEGMENT:     segment_command,
    LC_SYMTAB:      symtab_command,
    LC_LOAD_DYLIB:  dylib_command,
}

def getCommandClass(cmdtype):
    cls = command_classes.get(cmdtype)
    if cls != None:
        return cls
    return load_command




# CAN bus device mapping tool
from __future__ import print_function

import sys
import argparse
import re
import time
import signal
import importlib
import traceback

import cancat
from cancat.utils.types import SparseHexRange, ECUAddress
from cancat.utils import log
from cancat.uds.ecu import ECU
from cancat.uds.utils import ecu_did_scan, ecu_session_scan

c = None
_config = None
_output_filename = None
_can_session_filename = None


def now():
    # return an ISO 8601 formatted date string
    # 2019-07-18T02:28:16+00:00
    return time.strftime("%Y-%m-%dT%H:%M:%S%z", time.localtime())


class ECURange(SparseHexRange):
    def __new__(cls, val):
        # Ensure that values are <= than 0xFF (255)
        if any(int(n, 16) > 0xFF for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} exceeds max ECU addr of 0xFF'.format(val))
        return super(ECURange, cls).__new__(cls, val)


class DIDRange(SparseHexRange):
    def __new__(cls, val):
        # Ensure that values are <= than 0xFFFF (65535)
        if any(int(n, 16) > 0xFFFF for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} exceeds max DID of 0xFFFF'.format(val))
        return super(DIDRange, cls).__new__(cls, val)


class DiagnosticSessionRange(SparseHexRange):
    def __new__(cls, val):
        # Ensure that values are >= 2 and <= than 0x7F (127)
        if any(int(n, 16) > 0x7F for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} exceeds max Diagnostic Session of 0x7F'.format(val))
        elif any(int(n, 16) < 0x02 for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} below min Diagnostic Session of 0x02'.format(val))
        return super(DiagnosticSessionRange, cls).__new__(cls, val)


class SecurityAccessKeyRange(SparseHexRange):
    def __new__(cls, val):
        # Ensure that values are >= 2 and <= than 0x7D (125)
        if any(int(n, 16) > 0x7D for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} exceeds max Security Access Key of 0x7D'.format(val))
        elif any(int(n, 16) < 0x02 for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} below min Security Access Key of 0x02'.format(val))

        # The auth levels alternate
        return super(SecurityAccessKeyRange, cls).__new__(cls, val, 2)

class PayloadLength(SparseHexRange):
    def __new__(cls, val):
        # Ensure that values are <= than 0xFFF (4095)
        if any(int(n, 16) > 0xFFF for n in re.findall(r'[A-Za-z0-9]+', val)):
            raise ValueError('range {} exceeds max payload length of 0xFFF'.format(val))
        return super(PayloadLength, cls).__new__(cls, val)

class OneOrMoreOf(argparse.Action):
    def __init__(self, option_strings, dest, choices, nargs, **kwargs):
        if nargs == 0:
            raise ValueError('nargs must not be 0')
        if choices is None:
            raise ValueError('choices must be supplied')

        self.allowed_values = list(choices)

        # Override choices sent to the parent class constructor
        super(OneOrMoreOf, self).__init__(option_strings, dest, nargs=nargs, choices=None, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        # Handle both -sE -sD and -sED argument styles for the scan type
        items = getattr(namespace, self.dest, [])
        if items:
            items = list(items)
        else:
            items = []

        for val in values:
            if len(val) > 1:
                items.extend(list(val))
            else:
                items.append(val)

        # Get rid of duplicate entries
        items = list(set(items))

        if items and not set(items).issubset(set(self.allowed_values)):
            raise ValueError('invalid choice: {}, choose one or more of {}'.format(
                ''.join(items), ''.join(self.allowed_values)))
        setattr(namespace, self.dest, items)


def _get_baud_options():
    bauds = [b[4:-3] for b in dir(cancat) if b.startswith('CAN_') and b.endswith('BPS')]
    return bauds

def _get_baud_value(baud):
    ret = getattr(cancat, "CAN_{}BPS".format(baud))
    return ret


def udsmap_parse_args():
    # TODO: Add support for a writable did scan?
    # TODO: Add support to attempt memory read
    # TODO: Add support to attempt memory write?
    # TODO: Add support to attempt block data transfer from ECU?
    # TODO: Add support to attempt block data transfer to ECU?
    parser = argparse.ArgumentParser(
            prog='udsmap',
            description='CAN bus network mapping tool')
    parser.add_argument('-s', '--scan', required=True,
            #choices='EDWSAL', nargs='+', action=OneOrMoreOf,
            #help='Type of scan to run, select one or more of: (E) ECUs, (D) read DIDs, (W) write DIDs, (S) diagnostic Sessions, (A) seed/key Authentication levels, (L) authentication key Length')
            # re-enable only after security access levels and key length scanning has been tested
            #choices='EDSAL', nargs='+', action=OneOrMoreOf,
            #help='Type of scan to run, select one or more of: (E) ECUs, (D) read DIDs, (S) diagnostic Sessions, (A) security Authentication levels, (L) authentication key Length')
            choices='EDS', nargs='+', action=OneOrMoreOf,
            help='Type of scan to run, select one or more of: (E) ECUs, (D) read DIDs, (S) diagnostic Sessions')
    parser.add_argument('-p', '--port', default='/dev/ttyACM0',
            help='System device to use to communicate to the CanCat hardware (/dev/ttyACM0)') 
    parser.add_argument('-b', '--baud',
            choices=_get_baud_options(), default='AUTO',
            help='Set the CAN Bus Speed') 
    parser.add_argument('-t', '--discovery-type',
            choices=['did', 'session'], default='did',
            help='ECU discovery method: attempt to read a DID (F190), or enter diagnostic session 2')
    parser.add_argument('-m', '--bus-mode',
            choices=['std', 'ext', 'both'], default='both',
            help='Bus mode, only run standard 11-bit ECU discovery, extended 29-bit discovery or both - in the default "both" mode a standard 11-bit scan is run and then an exgtended 29-bit scan')
    parser.add_argument('-n', '--no-recursive-session-scanning', action='store_true',
            help='Disable recursive session scanning')
    parser.add_argument('-E', metavar='<ECU Range>',
            type=ECURange, default='00-FF',
            help='ECU address range to search')
    parser.add_argument('-D', metavar='<DID Read Range>',
            type=DIDRange, default='F180-F18E,F190-F1FF',
            help='DID range to search, by default restricted to the ISO14229 specified DIDs (F180-F18E,F190-F1FF), large DID ranges may take a long time')
    parser.add_argument('-S', metavar='<Diagnostic Session Range>',
            type=DiagnosticSessionRange, default='02-7F',
            help='Diagnostic Sessions to search for')
    #parser.add_argument('-A', metavar='<Security Access Key Range>',
    #        type=SecurityAccessKeyRange, default='01-41,61-7D',
    #        help='Security Access Key range to test, by default Security Sessions 43-60 are not tested, those values are used for end-of-life airbag deployment and may cause damage.')

    # According to ISO 14229 there is no upper bound on the size of the key 
    # (that I could see) aside from the constraints of the ISO-TP protocol 
    # itself. But for sanity's sake the default will be to stop trying when we 
    # reach 32 bytes.
    #parser.add_argument('-L', metavar='<Key Length Range>',
    #        type=PayloadLength, default='01-20',
    #        help='Key Length range to test')

    parser.add_argument('-T', '--timeout', type=float, default=0.2,
            help='UDS Timeout, 3 seconds is the ISO14229 standard, standard for this tool is 200 msec (0.2)')
    parser.add_argument('-w', '--startup-wait', type=float, nargs='?', const=2.0,
            help='Wait to receive CAN messages before starting the scan')
    parser.add_argument('-d', '--scan-delay', type=float, default=0.0,
            help='Wait a small time between requests, helps prevent flooding the bus')
    parser.add_argument('-r', '--rescan', action='store_true',
            help='Run a full rescan and merge new results with any existing data')
    #parser.add_argument('-f', '--force', action='store_true',
    #        help='Force udsmap to scan with potentially harmful options')
    #parser.add_argument('-y', '--yes', action='store_true',
    #        help='Automatically answer "yes" to any questions the tool asks')
    parser.add_argument('-v', '--verbose', action='count',
            help='Verbose logging, use -vv for extra verbosity')
    parser.add_argument('-l', '--log-file', nargs='?', const='canmap_%Y%m%d-%H%M%S.log',
            help='Log filename to write to, log filename can contain "time.strftime" formatting like "canmap_%%Y%%m%%d-%%H%%M%%S.log"')
    parser.add_argument('-o', '--output-file', nargs='?', const='canmap_%Y%m%d-%H%M%S.yml',
            help='Scan results output filename, can contain "time.strftime" formatting like "canmap_%%Y%%m%%d-%%H%%M%%S.yml"')
    parser.add_argument('-c', '--can-session-file', nargs='?', const='canmap_%Y%m%d-%H%M%S.sess',
            help='Filename for saving raw cancat session, can contain "time.strftime" formatting like "canmap_%%Y%%m%%d-%%H%%M%%S.sess"')
    parser.add_argument('-i', '--input-file',
            help='Input file containing previous scan results')
    parser.add_argument('-u', '--uds-class',
            help='Custom UDS class, allows for implementing key/seed unlock functions or testing, example: cancat.uds.test.TestUDS')
    return parser.parse_args()


def log_and_save(results, note):
    log.msg(note)
    results['notes'][results['start_time']] += '\n' + note


def import_results(args, c, scancls):
    if args.input_file is not None:
        # TODO: support multiple in/out file types?
        import yaml
        try:
            from yaml import CLoader as yamlLoader
        except ImportError:
            from yaml import Loader as yamlLoader

        with open(args.input_file, 'r') as f:
            imported_data = yaml.load(f, Loader=yamlLoader)

        config = {
            'config': imported_data['config'],
            'notes': imported_data['notes'],
            'ECUs': {},
        }

        if not config['notes']:
            config['notes'] = {}

        # If the input baudrate is AUTO (the default), and the input config
        # file has a baud rate, use the value from the config file, otherwise
        # override the config file.
        if args.baud != 'AUTO' or 'baud' not in config['config']:
            config['config']['baud'] = args.baud

        if 'no_recursive_session_scanning' not in config['config']:
            config['config']['no_recursive_session_scanning'] = args.no_recursive_session_scanning

        if 'timeout' not in config['config']:
            config['config']['timeout'] = args.timeout

        for e in imported_data['ECUs']:
            addr = ECUAddress(**e)
            config['ECUs'][addr] = ECU(c, addr, uds_class=scancls,
                    timeout=config['config']['timeout'], delay=args.scan_delay, **e)
        return config


def save_results(results, filename=None):
    if filename is not None:
        import yaml

        class literal_unicode(unicode): pass
        def literal_unicode_representer(dumper, data):
            return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')
        yaml.add_representer(literal_unicode, literal_unicode_representer)

        def hexint_presenter(dumper, data):
            return dumper.represent_int(hex(data))
        yaml.add_representer(int, hexint_presenter)

        def ecu_presenter(dumper, ecu):
            obj = []
            for key, val in ecu.export().items():
                obj.append((dumper.represent_data(key), dumper.represent_data(val)))
            return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', obj)

        yaml.add_representer(ECU, ecu_presenter)

        # TODO: It'd be more esthetically pleasing in the output files if 
        #       response data could be represented as a simple ascii string with 
        #       escaped bytes in it rather than the standard yaml "!!binary 
        #       + base64 string" method.
        #
        #       This is the closest I got but that makes it difficult to load 
        #       the data. For now we will use the default !!binary format, if 
        #       someone manually creates the data with escaped strings it'll 
        #       still be loaded correctly.
        #
        #       class data_str(str): pass
        #         def data_str_presenter(dumper, data):
        #             return dumper.represent_str(repr(data)[1:-1])
        #         yaml.add_representer(data_str, data_str_presenter)

        # TODO: I should probably make the config a class of it's own also, with 
        #       .add_ecu(), .add_note(), .export(), .import()
        output_data = {}
        global _config
        output_data['config'] = results['config']
        output_data['notes'] = dict([(k, literal_unicode(v)) for k, v in results['notes'].items()])
        output_data['ECUs'] = sorted([e for e in results['ECUs'].values()], key=lambda x: x._addr.tx_arbid)

        with open(filename, 'w') as f:
            f.write(yaml.dump(output_data))

def save():
    global _config, _output_filename, _can_session_filename
    if _output_filename:
        save_results(_config, _output_filename)

    if _can_session_filename:
        global c
        c.saveSessionToFile(_can_session_filename)


def save_and_exit(retval):
    save()
    sys.exit(retval)


def sigint_handler(signum, frame):
    global _config
    log_and_save(_config, 'scan aborted @ {}'.format(now()))
    save_and_exit(1)


def scan(config, args, c, scancls):
    if 'E' in args.scan:
        log_and_save(_config, 'ECU scan started @ {}'.format(config['start_time']))

        if not _config['ECUs'] or args.rescan:
            ecus = []
            if args.discovery_type == 'did':
                if args.bus_mode in ['std', 'both']:
                    ecus.extend(ecu_did_scan(c, args.E, ext=0, udscls=scancls,
                        timeout=config['config']['timeout'], delay=args.scan_delay))
                if args.bus_mode in ['ext', 'both']:
                    ecus.extend(ecu_did_scan(c, args.E, ext=1, udscls=scancls,
                        timeout=config['config']['timeout'], delay=args.scan_delay))
            else:
                if args.bus_mode in ['std', 'both']:
                    ecus.extend(ecu_session_scan(c, args.E, ext=0, udscls=scancls,
                        timeout=config['config']['timeout'], delay=args.scan_delay))
                if args.bus_mode in ['ext', 'both']:
                    ecus.extend(ecu_session_scan(c, args.E, ext=1, udscls=scancls,
                        timeout=config['config']['timeout'], delay=args.scan_delay))

            for addr in ecus:
                _config['ECUs'][addr] = ECU(c, addr, uds_class=scancls,
                        timeout=config['config']['timeout'], delay=args.scan_delay)

    if 'D' in args.scan:
        log_and_save(_config, 'DID read scan started @ {}'.format(now()))

        for ecu in _config['ECUs'].values():
            ecu.did_read_scan(args.D, args.rescan)

    #if 'W' in args.scan:
    #    log_and_save(_config, 'DID write scan started @ {}'.format(now()))
    #
    #    for ecu in _config['ECUs'].values():
    #        ecu.did_write_scan(args.D, args.rescan)

    if 'S' in args.scan:
        log_and_save(_config, 'Session scan started @ {}'.format(now()))

        if config['config']['no_recursive_session_scanning']:
            recursive = False
        else:
            recursive = True

        for ecu in _config['ECUs'].values():
            ecu.session_scan(args.S, args.rescan, rescan_did_range=args.D, recursive_scan=recursive)

    if 'A' in args.scan:
        log_and_save(_config, 'Auth scan started @ {}'.format(now()))

        for ecu in _config['ECUs'].values():
            ecu.auth_scan(args.A, args.rescan)

    if 'L' in args.scan:
        log_and_save(_config, 'Key Length scan started @ {}'.format(now()))

        for ecu in _config['ECUs'].values():
            ecu.key_length_scan(args.L, args.rescan)


def main():
    args = udsmap_parse_args()
    
    # TODO: move this to it's own class in cancat.utils, the config/args global 
    #       data thing will be better handled that way
    #cancat.utils.canmap(args)

    if args.verbose == 1:
        loglevel = log.DEBUG
    elif args.verbose >= 2:
        loglevel = log.DETAIL
    else:
        loglevel = log.WARNING

    if args.log_file is not None:
        logfile = time.strftime(args.log_file)
    else:
        logfile = None

    log.start(level=loglevel, filename=logfile)
    log.debug(args)

    # TODO: Check for potentially damaging options
    
    # TODO: Warn about how long the combined scan options will probably take:
    #   - ECU Scan: if timeout is 3s:
    #                   worst case with both std & ext = (247 + 255) * 3 =
    #                   3012 sec = 25.1 minutes
    #   - ECU Scan: if timeout is 0.1s:
    #                   worst case with both std & ext = (247 + 255) * 0.1 =
    #                   50 sec = ~ 1 minute
    #   - DID Scan: 
    #log.warning('ECU scan {} may take up to 25 minutes to complete'.format(args.E))

    global c

    ifaceclass = cancat.CanInterface
    scancls = None
    if args.uds_class:
        pkg, cls = args.uds_class.rsplit('.', 1)
        scanlib = importlib.import_module(pkg)
        scancls = getattr(scanlib, cls)

        # If the custom UDS class package has a "CanInterface" class, use that 
        # instead of the real cancat.CanInterface class
        if hasattr(scanlib, 'CanInterface'):
            ifaceclass = getattr(scanlib, 'CanInterface')
    c = ifaceclass(port=args.port)

    global _config
    if args.input_file is not None:
        _config = import_results(args, c, scancls)
    else:
        _config = {
            'config': {
                'baud': args.baud,
                'timeout': args.timeout,
                'no_recursive_session_scanning': args.no_recursive_session_scanning,
            },
            'notes': {},
            'ECUs': {},
        }

    c.setCanBaud(_get_baud_value(_config['config']['baud']))

    if args.output_file:
        global _output_filename
        _output_filename = time.strftime(args.output_file)

    if args.can_session_file:
        global _can_session_filename
        _can_session_filename = time.strftime(args.can_session_file)

    start_time = now()
    _config['notes'][start_time] = []
    _config['start_time'] = start_time
    _config['notes'][start_time] = 'command: {}'.format(' '.join(sys.argv))

    if args.startup_wait:
        # Listen for messages to ensure that the bus is working right
        count1 = c.getCanMsgCount()
        time.sleep(args.startup_wait)
        count2 = c.getCanMsgCount()
        if count2 <= count1:
            log_and_save(_config, 'ERROR: No CAN traffic detected on {} @ {}'.format(args.port, args.baud))
            save_and_exit(2)
    
    # signal will catch CTRL-C in script contexts
    signal.signal(signal.SIGINT, sigint_handler)

    # catching KeyboardInterrupt will catch CTRL-C in interactive contexts
    try:
        scan(_config, args, c, scancls)

        log_and_save(_config, 'scans completed @ {}'.format(now()))
        save_and_exit(0)
    except KeyboardInterrupt:
        log_and_save(_config, 'scan aborted @ {}'.format(now()))
        save_and_exit(1)
    except Exception as e:
        saved_trace = traceback.format_exc()
        log_and_save(_config, 'Exception @ {}: {}'.format(now(), e))
        save()

        print(saved_trace, file=sys.stderr)
        sys.exit(3)



# Utility functions for UDS

import time
import string
import struct
from contextlib import contextmanager

from cancat import uds
from cancat.uds import UDS
from cancat.utils import log
from cancat.utils.types import ECUAddress, _range_func


def get_uds_29bit_srcid(arbid):
    consts = uds.ARBID_CONSTS['29bit']
    return arbid & consts['srcid_mask']


def get_uds_29bit_destid(arbid):
    consts = uds.ARBID_CONSTS['29bit']
    return (arbid & consts['destid_mask']) >> consts['destid_shift']


def gen_uds_resp_range(arbid):
    if arbid > uds.ARBID_CONSTS['29bit']['prefix']:
        # Normally if a request is sent to 0x18DA01F1, the response should have 
        # an arbitration ID of 0x18DAF101, but not all ECUs do things in 
        # a "normal" way, so generate a range of possible response IDs.

        # The src from the request will be come the destination in the response
        dest_id = get_uds_29bit_srcid(arbid)
        base_id = uds.ARBID_CONSTS['29bit']['prefix'] & (dest_id << uds.ARBID_CONSTS['29bit']['destid_shift'])

        return _range_func(base_id, base_id + 0x100)
    else:
        # Assume this is an 11-bit request
        #
        # Normally if a request is sent to 0x710, the response should have an 
        # arbitration ID of 0x718, but not all ECUs do things in a "normal" way, 
        # so generate a range of possible response IDs.
        return _range_func(0x700, 0x800)


def gen_arbids(idx, ext=0):
    if ext:
        prefix = uds.ARBID_CONSTS['29bit']['prefix']
        tester = uds.ARBID_CONSTS['29bit']['tester']
        dest_shift = uds.ARBID_CONSTS['29bit']['destid_shift']

        arb_id = prefix + (idx << dest_shift) + tester
        resp_id = prefix + (tester << dest_shift) + idx
    else:
        prefix = uds.ARBID_CONSTS['11bit']['prefix']

        arb_id = prefix + idx
        resp_id = prefix + idx + uds.ARBID_CONSTS['11bit']['resp_offset']

    return (arb_id, resp_id)


def enter_session(u, session, prereq_sessions=None):
    # Enter any required preq sessions
    if prereq_sessions:
        for prereq in prereq_sessions:
            #u.DiagnosticSessionControl(prereq)
            enter_session(u, prereq)

    return u.DiagnosticSessionControl(session)

@contextmanager
def new_session(u, session, prereq_sessions=None, tester_present=False):
    try:
        # Enter any required preq sessions
        if prereq_sessions:
            for prereq in prereq_sessions:
                #u.DiagnosticSessionControl(prereq)
                enter_session(u, prereq)

        msg = u.DiagnosticSessionControl(session)

        # Start tester present again
        if tester_present:
            u.StartTesterPresent(request_response=False)

        yield msg
    finally:
        if tester_present:
            u.StopTesterPresent()


def find_possible_resp(u, start_index, tx_arbid, service, subfunction=None, timeout=3.0):
    # Starting at the supplied starting index, find the service request, and
    # then look for possible responses until the supplied timeout
    
    if subfunction:
        tx_match_bytes = struct.pack('>BH', service, subfunction)
        rx_match_bytes = struct.pack('>BH', service + 0x40, subfunction)
        match_len = 3
    else:
        tx_match_bytes = struct.pack('>B', service)
        rx_match_bytes = struct.pack('>B', service + 0x40)
        match_len = 1

    for idx, _, _, msg in u.c.genCanMsgs(start=start_index, arbids=[tx_arbid], maxsecs=timeout):
        ftype = msg[0] >> 4
        # Our Tx arbid is probably a 0, but check for frame type 1 as well.
        if (ftype == 0 and msg[1:1+match_len] == tx_match_bytes) or \
                (ftype == 1 and msg[2:2+match_len] == tx_match_bytes):
            tx_index = idx
            tx_msg = msg
            break
    else:
        # error case handling, if 0 messages are generated
        tx_index = 0
        tx_msg = None

    rx_range = gen_uds_resp_range(tx_arbid)
    err_match = struct.pack('>BB', uds.SVC_NEGATIVE_RESPONSE, service)

    for idx, _, arbid, msg in u.c.genCanMsgs(start=tx_index+1, arbids=rx_range, maxsecs=timeout):
        ftype = msg[0] >> 4
        # Check for frame types 0 (positive and negative responses) and 1
        if (ftype == 0 and msg[1:1+match_len] == rx_match_bytes) or \
                (ftype == 0 and msg[1:3] == err_match) or \
                (ftype == 1 and msg[2:2+match_len] == rx_match_bytes):
            return tx_msg, (arbid, msg)

    return tx_msg, None 


def err_str(err):
    try:
        return uds.NEG_RESP_CODES.get(err)
    except IndexError:
        return 'ISOSAEReserved({})'.format(hex(err))


def did_str(did):
    name_str = uds.ISO_14229_DIDS.get(did)

    # Handle some generic DID ranges
    if name_str is None:
        if did in _range_func(0xf1a0, 0xf1ef+1):
            name_str = 'identificationOptionVehicleManufacturerSpecific'
        elif did in _range_func(0xf1f0, 0xf1ff+1):
            name_str = 'identificationOptionSystemSupplierSpecific'

    if name_str:
        return '{} ({})'.format(hex(did), name_str)
    else:
        return hex(did)


def ecu_did_scan(c, arb_id_range, ext=0, did=0xf190, udscls=None, timeout=3.0, delay=None, verbose_flag=False):
    scan_type = ''
    if ext:
        scan_type = ' ext'

    if udscls is None:
        udscls = UDS

    log.debug('Starting{} DID read ECU scan for range: {}'.format(scan_type, arb_id_range))
    c.placeCanBookmark('ecu_did_scan({}, ext={}, did={}, timeout={}, delay={})'.format(
        arb_id_range, ext, did, timeout, delay))
    ecus = []
    possible_ecus = []
    for i in arb_id_range:  
        if ext and i == uds.ARBID_CONSTS['29bit']['tester']:
            # Skip i == 0xF1 because in that case the sender and receiver IDs 
            # are the same
            log.detail('Skipping 0xF1 in ext ECU scan: invalid ECU address')
            continue
        elif ext == False and i > uds.ARBID_CONSTS['11bit']['max_req_id']:
            # For non-extended scans the valid range goes from 0x00 to 0xFF, but 
            # stop the scan at 0xf7 because at that time the response is the 
            # largest possible valid value
            log.detail('Stopping std ECU scan at 0xF7: last valid ECU address')
            break

        arb_id, resp_id = gen_arbids(i, ext)

        addr = ECUAddress(arb_id, resp_id, ext)

        u = udscls(c, addr.tx_arbid, addr.rx_arbid, extflag=addr.extflag,
                verbose=verbose_flag, timeout=timeout)
        log.detail('Trying {}'.format(addr))
        try:
            start_index = u.c.getCanMsgCount()
            msg = u.ReadDID(did)
            if msg is not None:
                log.debug('{} DID {}: {}'.format(addr, hex(did), repr(msg)))
                log.msg('found {}'.format(addr))

                ecus.append(addr)
            else:
                tx_msg, possible_match = find_possible_resp(u, start_index, arb_id,
                        uds.SVC_READ_DATA_BY_IDENTIFIER, did, timeout)
                if possible_match:
                    log.warn('Possible non-standard responses for {} found:'.format(addr))
                    rx_arbid, msg = possible_match
                    log.warn('{}: {}'.format(hex(rx_arbid), msg.hex()))
                    possible_ecus.append(ECUAddress(arb_id, rx_arbid, ext))
        except uds.NegativeResponseException as e:
            log.debug('{} DID {}: {}'.format(addr, hex(did), e))
            log.msg('found {}'.format(addr))

            # If a negative response happened, that means an ECU is present 
            # to respond at this address.
            ecus.append(addr)

        if delay:
            time.sleep(delay)

    # Double check any non-standard responses that were found
    for addr in possible_ecus:  
        u = udscls(c, addr.tx_arbid, addr.rx_arbid, extflag=addr.extflag,
                verbose=verbose_flag, timeout=timeout)
        log.detail('Trying {}'.format(addr))
        try:
            msg = u.ReadDID(did)
            if msg is not None:
                log.debug('{} DID {}: {}'.format(addr, hex(did), repr(msg)))
                log.msg('found {}'.format(addr))
                ecus.append(addr)
        except uds.NegativeResponseException as e:
            log.debug('{} DID {}: {}'.format(addr, hex(did), e))
            log.msg('found {}'.format(addr))

            # If a negative response happened, that means an ECU is present 
            # to respond at this address.
            ecus.append(addr)

        if delay:
            time.sleep(delay)

    return ecus


def ecu_session_scan(c, arb_id_range, ext=0, session=1, udscls=None, timeout=3.0, delay=None, verbose_flag=False):
    scan_type = ''
    if ext:
        scan_type = ' ext'

    if udscls is None:
        udscls = UDS

    log.debug('Starting{} Session ECU scan for range: {}'.format(scan_type, arb_id_range))
    c.placeCanBookmark('ecu_session_scan({}, ext={}, session={}, timeout={}, delay={})'.format(
        arb_id_range, ext, session, timeout, delay))

    ecus = []
    possible_ecus = []
    for i in arb_id_range:  
        if ext and i == uds.ARBID_CONSTS['29bit']['tester']:
            # Skip i == 0xF1 because in that case the sender and receiver IDs 
            # are the same
            log.detail('Skipping 0xF1 in ext ECU scan: invalid ECU address')
            continue
        elif ext == False and i > uds.ARBID_CONSTS['11bit']['max_req_id']:
            # For non-extended scans the valid range goes from 0x00 to 0xFF, but 
            # stop the scan at 0xf7 because at that time the response is the 
            # largest possible valid value
            log.detail('Stopping std ECU scan at 0xF7: last valid ECU address')
            break

        arb_id, resp_id = gen_arbids(i, ext)

        addr = ECUAddress(arb_id, resp_id, ext)
        u = udscls(c, addr.tx_arbid, addr.rx_arbid, extflag=addr.extflag, verbose=verbose_flag, timeout=timeout)
        log.detail('Trying {}'.format(addr))
        try:
            start_index = u.c.getCanMsgCount()
            with new_session(u, session) as msg:
                if msg is not None:
                    log.debug('{} session {}: {}'.format(addr, session, repr(msg)))
                    log.msg('found {}'.format(addr))

                    ecus.append(addr)
                else:
                    tx_msg, responses = find_possible_resp(u, start_index, arb_id,
                            uds.SVC_DIAGNOSTICS_SESSION_CONTROL, session, timeout)
                    if responses:
                        log.warn('Possible non-standard responses for {} found:'.format(addr))
                        rx_arbid, msg = possible_match
                        log.warn('{}: {}'.format(hex(rx_arbid), msg.hex()))
                        possible_ecus.append(ECUAddress(arb_id, rx_arbid, ext))
        except uds.NegativeResponseException as e:
            log.debug('{} session {}: {}'.format(addr, session, e))
            log.msg('found {}'.format(addr))

            # If a negative response happened, that means an ECU is present 
            # to respond at this address.
            ecus.append(addr)

        if delay:
            time.sleep(delay)

    # Double check any non-standard responses that were found
    for addr in possible_ecus:  
        u = udscls(c, addr.tx_arbid, addr.rx_arbid, extflag=addr.extflag,
                verbose=verbose_flag, timeout=timeout)
        log.detail('Trying {}'.format(addr))
        try:
            with new_session(u, sess) as msg:
                if msg is not None:
                    log.debug('{} session {}: {}'.format(addr, sess, repr(msg)))
                    log.msg('found {}'.format(addr))
                    ecus.append(addr)
        except uds.NegativeResponseException as e:
            log.debug('{} session {}: {}'.format(addr, sess, e))
            log.msg('found {}'.format(addr))

            # If a negative response happened, that means an ECU is present 
            # to respond at this address.
            ecus.append(addr)

        if delay:
            time.sleep(delay)

    return ecus


def try_read_did(u, did):
    data = None
    try:
        resp = u.ReadDID(did)
        if resp is not None:
            data = { 'resp':resp }
    except uds.NegativeResponseException as e:
        # 0x12:'SubFunctionNotSupported' - not standard, but I've seen it
        # 0x31:'RequestOutOfRange'       - means the DID is not valid
        if e.neg_code not in [0x12, 0x31]:
            data = { 'err':e.neg_code }
    return data


def did_read_scan(u, did_range, delay=None):
    log.debug('Starting DID read scan for range: {}'.format(did_range))
    u.c.placeCanBookmark('did_read_scan({}, delay={})'.format(did_range, delay))
    dids = {}
    for i in did_range:  
        log.detail('Trying DID read {}'.format(hex(i)))
        u.c.placeCanBookmark('ReadDID({})'.format(hex(i)))
        resp = try_read_did(u, i)

        if resp is not None:
            log.debug('DID {}: {}'.format(hex(i), resp))
            if 'resp' in resp:
                printable_did = ''.join([chr(x) if chr(x) in string.printable else '' for x in resp['resp'][3:]])
                log.msg('DID {}: {} ({})'.format(did_str(i), resp['resp'], printable_did))
            else:
                log.msg('DID {}: ERR {}'.format(did_str(i), err_str(resp['err'])))
            dids[i] = resp

        if delay:
            time.sleep(delay)

    return dids


def try_write_did(u, did, datg):
    did = None
    try:
        resp = u.WriteDID(i, data)
        if resp is not None:
            did = { 'resp':resp }
    except uds.NegativeResponseException as e:
        # 0x31:'RequestOutOfRange' usually means the DID is not valid
        if e.neg_code != 0x31:
            did = { 'err':e.neg_code }
    return did


def did_write_scan(u, did_range, write_data, delay=None):
    log.debug('Starting DID write scan for range: {}'.format(did_range))
    u.c.placeCanBookmark('did_write_scan({}, write_data={}, delay={})'.format(did_range, write_data, delay))
    dids = {}
    for i in did_range:  
        log.detail('Trying DID write {}'.format(hex(i)))
        u.c.placeCanBookmark('WriteDID({})'.format(hex(i)))
        resp = try_write_did(u, i, write_data)
        if resp is not None:
            log.detail('DID {}: {}'.format(hex(i), resp))
            if 'resp' in resp:
                log.msg('DID {}: {}'.format(did_str(i), resp['resp'].hex()))
            else:
                log.msg('DID {}: {}'.format(did_str(i), err_str(resp['err'])))
            dids[i] = resp

        if delay:
            time.sleep(delay)

    return dids


def try_session(u, sess_num):
    session = None
    try:
        with new_session(u, sess_num) as resp:
            if resp is not None:
                session = { 'resp':resp }
    except uds.NegativeResponseException as e:
        # 0x11:'ServiceNotSupported'
        # 0x12:'SubFunctionNotSupported'
        if e.neg_code not in [0x11, 0x12]:
            session = { 'err':e.neg_code }
    return session

def try_session_scan(u, session_range, prereq_sessions, found_sessions, delay=None, recursive_scan=True, try_ecu_reset=True, try_sess_ctrl_reset=True):
    log.debug('Starting session scan for range: {}'.format(session_range))
    u.c.placeCanBookmark('session_scan({}, delay={})'.format(session_range, delay))
    sessions = {}
    for i in session_range:  
        # If this session matches any one of the prereqs, or one of the
        # sessions already found, skip it
        if i in prereq_sessions or i in found_sessions:
            continue

        log.detail('Trying session {}'.format(i))
        u.c.placeCanBookmark('DiagnosticSessionControl({})'.format(i))

        # Enter any required preq sessions
        try:
            for prereq in prereq_sessions:
                enter_session(u, prereq)
        except uds.NegativeResponseException as e:
            # 0x7f:'ServiceNotSupportedInActiveSession'
            if e.neg_code == 0x7f:
                log.detail('SESSION ({}): Can\'t enter prereqs, stopping session scan'.format(prereq_sessions))
                return sessions
            else:
                raise e

        resp = try_session(u, i)
        if resp is not None:
            resp['prereqs'] = list(prereq_sessions)
            log.debug('session {}: {}'.format(i, resp))
            if 'resp' in resp:
                log.msg('SESSION {}: {} ({})'.format(i, resp['resp'].hex(), prereq_sessions))
            else:
                log.msg('SESSION {}: {} ({})'.format(i, err_str(resp['err']), prereq_sessions))
            sessions[i] = resp

        # Only bother with this if a successful response was received
        if resp and 'resp' in resp:
            if try_ecu_reset:
                try:
                    u.EcuReset()

                    # Small delay to allow for the reset to complete
                    time.sleep(0.2)
                except uds.NegativeResponseException as e:
                    # 0x11:'ServiceNotSupported'
                    # 0x22:'ConditionsNotCorrect'
                    if e.neg_code in [0x11, 0x22]:
                        try_ecu_reset = False
            elif try_sess_ctrl_reset:
                try:
                    # Try just changing back to session 1
                    new_session(u, 1)
                except uds.NegativeResponseException as e:
                    # The default method to try returning to session 1 is EcuReset, if
                    # EcuReset doesn't work (or isn't enabled), then try using the
                    # DiagnosticSessionControl message to return to session 1, if that
                    # doesn't work then we can't attempt recursive session scanning
                    try_sess_ctrl_reset = False

        # Extra delay between attempts if configured
        if delay:
            time.sleep(delay)

    # For each session found re-scan for new sessions that can be entered from
    # those, but only if we have a valid reset method:
    if recursive_scan and (try_ecu_reset or try_sess_ctrl_reset):
        subsessions = {}
        for sess in sessions:
            # Only attempt this with sessions that we got a successful response 
            # for
            if 'msg' in sessions[sess]:
                log.debug('Scanning for sessions from session {} ({})'.format(sess, prereq_sessions))
                prereqs = prereq_sessions + [sess]
                subsessions.update(try_session_scan(u, session_range, prereqs, sessions.keys(),
                    delay=delay, recursive_scan=recursive_scan,
                    try_ecu_reset=try_ecu_reset, try_sess_ctrl_reset=try_sess_ctrl_reset))
        sessions.update(subsessions)

    return sessions


def session_scan(u, session_range, delay=None, recursive_scan=True):
    prereq_sessions = []
    session_results = try_session_scan(u, session_range, prereq_sessions, [], delay, recursive_scan)
    return session_results


def try_auth(u, level, key):
    auth_data = None
    try:
        resp = u.SecurityAccess(level, key)
        if resp is not None:
            auth_data = { 'resp':resp }
    except uds.NegativeResponseException as e:
        # 0x12:'SubFunctionNotSupported',
        if e.neg_code != 0x12:
            auth_data = { 'err':e.neg_code }
    return auth_data


def auth_scan(u, auth_range, key_func=None, delay=None):
    log.debug('Starting auth scan for range: {}'.format(auth_range))
    u.c.placeCanBookmark('auth_scan({}, key_func={}, delay={})'.format(auth_range, key_func, delay))
    auth_levels = {}
    for i in auth_range:  
        if key_func:
            key = ''
        else:
            key = key_func(i)

        log.detail('Trying auth level {}: key \'{}\''.format(i, key))
        u.c.placeCanBookmark('SecurityAccess({}, {})'.format(i, repr(key)))

        resp = try_auth(u, i, key)
        if resp is not None:
            log.debug('auth {}: {}'.format(i, resp))
            if 'resp' in resp:
                log.msg('SECURITY {}: {}'.format(i, resp['resp'].hex()))
            else:
                log.msg('SECURITY {}: {}'.format(i, err_str(resp['err'])))
            auth_levels[i] = resp

        if delay:
            time.sleep(delay)

    return auth_levels



# Command line entry point for cancat2candump

import sys
import argparse
from cancat.utils import convert


def main():
    argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
            prog='cancat2candump',
            description='Utility to convert a CanCat session into a candump log')
    parser.add_argument('session', help='input CanCat session')
    parser.add_argument('output', help='output candump file')
    args = parser.parse_args(argv)

    convert.cancat2candump(args.session, args.output)



# Command line entry point for cancat2pcap
# This requires scapy to be installed

import sys
import argparse
from cancat.utils import convert


def main():
    argv = sys.argv[1:]
                                                                           
    parser = argparse.ArgumentParser(
            prog='cancat2pcap',
            description='Utility to convert a CanCat session into a pcap')
    parser.add_argument('session', type=argparse.FileType('rb'),
            help='input CanCat session')
    parser.add_argument('output', type=str,
            help='output pcap file')
    args = parser.parse_args(argv)

    convert.cancat2pcap(args.session, args.output)



'''
Structure definitions for the OSX MachO binary format.
'''
from __future__ import print_function
from past.builtins import xrange

import struct
import vstruct

from vstruct.defs.macho.fat import *
from vstruct.defs.macho.const import *
from vstruct.defs.macho.stabs import *
from vstruct.defs.macho.loader import *

class mach_o(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self._raw_bytes = ''
        self._symbols = None

        self.mach_header = mach_header()
        self.load_commands = vstruct.VStruct()

    def getSymbols(self):
        if self._symbols != None:
            return self._symbols

        self._symbols = []
        for fname,vs in self.load_commands:
            if vs.cmd != LC_SYMTAB:
                continue
            strbytes = self._raw_bytes[vs.stroff:vs.stroff+vs.strsize]
            #print(repr(strbytes))
            strtab = strbytes.split('\x00')
            #print(strtab)
            offset = vs.symoff
            for i in xrange(vs.nsyms):
                n = nlist() # FIXME 64!
                offset = n.vsParse(self._raw_bytes, offset)
                #symstr = strtab[n.n_strx]
                # FIXME this is slow!
                symstr = strbytes[n.n_strx:].split('\x00', 1)[0]
                #print(n.tree())
                #print(symstr)

    def getLibDeps(self):
        '''
        Return a list of the library files this Mach-O is dependant on
        '''
        ret = []
        for fname, vs in self.load_commands:
            if vs.cmd != LC_LOAD_DYLIB:
                continue
            ret.append(vs.namedata)
        return ret

    def getSegments(self):
        '''
        Return a list of (segname, rva, perms, bytes) tuples for the memory
        segments defined by the loader commands
        '''
        ret = []
        for fname, vs in self.load_commands:
            if vs.cmd != LC_SEGMENT:
                #print(hex(vs.cmd),hex(vs.cmdsize) # 2, 5, b, e)
                continue
            # Slice the segment bytes from raw bytes
            fbytes = self._raw_bytes[ vs.fileoff: vs.fileoff + vs.filesize ]
            # Pad out to virtual size
            fbytes = fbytes.ljust(vs.vmsize, '\x00')

            ret.append((vs.segname, vs.vmaddr, vs.initprot, fbytes))
        return ret

    def vsParse(self, bytes, offset=0):
        self._raw_bytes = bytes[offset:]
        offset = self.mach_header.vsParse(bytes, offset=offset)
        #print(bytes[offset:].encode('hex'))
        for i in xrange(self.mach_header.ncmds):
            # should we use endian from header?
            cmdtype, cmdlen = struct.unpack('<II', bytes[offset:offset+8])
            cmdclass = getCommandClass(cmdtype)
            cmdobj = cmdclass()
            cmdobj.vsParse(bytes, offset=offset)
            self.load_commands.vsAddField('cmd%d' % i, cmdobj)
            offset += cmdobj.cmdsize




import unittest

from cancat.utils.types import ECUAddress
from cancat.uds.ecu import ECU
from cancat.uds.utils import ecu_did_scan, ecu_session_scan
from cancat.uds.test import CanInterface, TestUDS

class UDStest(unittest.TestCase):
    def test_ecu_did_scan(self):
        c = CanInterface('')
        testclass_ecus = [
            ECUAddress(0x711, 0x719, 0),
            ECUAddress(0x7e0, 0x7e8, 0)
        ]

        ecus = ecu_did_scan(c, range(0, 0x100), udscls=TestUDS)
        self.assertEqual(testclass_ecus, ecus)

    def test_ecu_session_scan(self):
        c = CanInterface('')
        testclass_ecus = [
            ECUAddress(0x711, 0x719, 0),
            ECUAddress(0x7e0, 0x7e8, 0)
        ]

        ecus = ecu_session_scan(c, range(0, 0x100), udscls=TestUDS)
        self.assertEqual(testclass_ecus, ecus)

    def test_did_scan(self):
        c = CanInterface('')
        testclass_ecus = [
            {
                'ecu': ECU(c, ECUAddress(0x711, 0x719, 0), scancls=TestUDS),
                'dids': {
                    0x0042: '\x62\x00\x42ANSWER',
                },
            },
            {
                'ecu': ECU(c, ECUAddress(0x7e0, 0x7e8, 0), scancls=TestUDS),
                'dids': {
                    0xE010: '\x62\xE0\x10VERSION 1.2.3',
                    0xF190: '\x62\xF1\x901AB123CD1EF123456',
                },
            },
        ]
        for test in testclass_ecus:
            ecu = test['ecu']
            test_dids = list(test['dids'].keys())

            before_scan_sessions = list(ecu._sessions.keys())
            self.assertEqual([1], before_scan_sessions)

            before_scan_dids = list(ecu._sessions[1]['dids'].keys())
            self.assertEqual([], before_scan_dids)

            ecu.did_read_scan(range(0, 0x10000))

            after_scan_sessions = list(ecu._sessions.keys())
            self.assertEqual([1], after_scan_sessions)

            dids = list(ecu._sessions[1]['dids'].keys())
            self.assertEqual(test_dids, dids)

            for did in test['dids'].keys():
                self.assertEqual(test['dids'][did], ecu._sessions[1]['dids'][did]['resp'])




import usb
import time
import queue
import logging
import unittest
import threading
import traceback

from rflib.const import *
from rflib.bits import ord23

logging.basicConfig(level=logging.INFO, format='%(asctime)s:%(levelname)s:%(name)s: %(message)s')
logger = logging.getLogger(__name__)

EP0BUFSIZE = 512


class fakeMemory:
    def __init__(self, size=64*1024):
        self.memory = [0 for x in range(size)]
        self.mmio = {}

    def readMemory(self, addr, size):
        logger.debug("fm.readMemory(0x%x, 0x%x)", addr, size)
        chunk = b''.join([b'%c' % x for x in self.memory[addr:addr+size]])
        if len(chunk) < size:
            chunk += b"@" * (size-len(chunk))
        return chunk

    def writeMemory(self, addr, data):
        logger.debug("fm.writeMemory(0x%x, %r)", addr, data)
        #if type(data) == str:
        #    raise(Exception("Cannot write 'str' to fakeMemory!  Must use 'bytes'"))

        for x in range(len(data)):
            tgt = addr+x
            val = data[x]

            handler = self.mmio.get(tgt)
            if handler is not None:
                val = handler(tgt, data[x])

            # if we didn't return None from the handler, write it anyway
            if val is not None:
                self.memory[tgt] = val

    '''
    def mmio_RFST(self, tgt, dbyte):
        logger.info('mmio_RFST(0x%x, %r)', tgt, dbyte)
        print("RFST==%x  (%x)" % (self.readMemory(X_RFST, 1), ord(dbyte)))


        # configure MARCSTATE
        val = ord(dbyte)
        if val in (2, 3):
            val = dbyte+10

        else:
            val = MARC_STATE_RX

        self.writeMemory(MARCSTATE, b'%c'%(val))

        # still set RFST
        return dbyte

    def mmio_MARCSTATE(self, tgt, dbyte):
        rfst = self.readMemory(X_RFST, 1)
        logger.info('mmio_MARCSTATE(0x%x, %r) rfst=%r', tgt, dbyte, rfst)
        return MARC_STATE_RX
    '''

## Serial Commands ##
CMD_LOG                 = 0x2f
CMD_LOG_HEX             = 0x2e

CMD_CAN_RECV                   = 0x30
CMD_PING_RESPONSE              = 0x31
CMD_CHANGE_BAUD_RESULT         = 0x32
CMD_CAN_BAUD_RESULT            = 0x33
CMD_CAN_SEND_RESULT            = 0x34
CMD_ISO_RECV                   = 0x35
CMD_SET_FILT_MASK              = 0x36
CMD_CAN_MODE_RESULT            = 0x37
CMD_CAN_SEND_ISOTP_RESULT      = 0x38
CMD_CAN_RECV_ISOTP_RESULT      = 0x39
CMD_CAN_SENDRECV_ISOTP_RESULT  = 0x3A
CMD_PRINT_CAN_REGS             = 0x3C

CMD_PING                = 0x41
CMD_CHANGE_BAUD         = 0x42
CMD_CAN_BAUD            = 0x43
CMD_CAN_SEND            = 0x44
CMD_CAN_MODE            = 0x45
CMD_CAN_MODE_SNIFF_CAN0 = 0x00
CMD_CAN_MODE_SNIFF_CAN1 = 0x01
CMD_CAN_MODE_CITM       = 0x02
CMD_CAN_SEND_ISOTP      = 0x46
CMD_CAN_RECV_ISOTP      = 0x47
CMD_CAN_SENDRECV_ISOTP  = 0x48

# TODO: first, implement PING capability



class FakeCanCat:
    '''
    This class emulates a real CanCat (the physical device).
    We're going to try making this single-threaded.  We may need to handle in/out in separate threads, but fingers crossed.
    '''
    def __init__(self):
        self._inbuf = b''
        self._outbuf = b''

        self._inq = queue.Queue()
        self._outq = queue.Queue()  # this is what's handed to the CanCat receiver thread
        self.memory = fakeMemory()

        self.start_ts = time.time()

        #self.memory.writeMemory(0xdf00, FAKE_MEM_DF00)
        #self.memory.writeMemory(0xdf46, b'\xf0\x0d')
        #for intreg, intval in list(FAKE_INTERRUPT_REGISTERS.items()):
        #    logger.info('setting interrupt register: %r = %r', intreg, intval)
        #    self.memory.writeMemory(eval(intreg), intval)

        # do we want to add in a few CAN messages to be received?

    def clock(self):
        return time.time() - self.start_ts

    def CanCat_send(self, data, cmd):
        print(b'===FakeCanCat_send: cmd:%x data: %r' % (cmd, data))
        packet = b'@%c%c%s' % (len(data)+1, cmd, data)
        self._inq.put(packet)

    def log(self, msg):
        self.CanCat_send(b"FakeCanCat: " + msg, CMD_LOG)
    def logHex(self, num):
        self.CanCat_send(struct.pack(b"<I", num), CMD_LOG_HEX)
    def logHexStr(self, num, prefix):
        self.log(prefix)
        self.logHex(num)

    #### FAKE SERIAL DEVICE (interface to Python)
    def read(self, count=1):
        if len(self._inbuf) < count:
            empty = False
            try:
                #print(b'===fccc: attempting to get from _inq:')
                self._inbuf += self._inq.get(timeout = 1)
            except queue.Empty:
                empty = True
                #print(b'===fccc: attempting to get from _inq:  NOPE')
                return b''

        out = self._inbuf[:count]
        self._inbuf = self._inbuf[count:]
        if len(out):
            #print(b"===FakeCanCatCHEAT===: read(%r):  %x" % (count, ord(out)))
            return out

        return b''

    def write(self, msg):
        #self._inq.put(msg) # nah, let's try to handle it here...

        self.log(b"write(%r)" % msg)
        print(b"===FakeCanCatCHEAT===: write(%r)" % msg)

        length, cmd = struct.unpack("<HB", msg[0:3])
        data = msg[3:]

        if cmd == CMD_CHANGE_BAUD:
            print(b'=CMD_CHANGE_BAUD=')
            self.log(b"CMD_CHANGE_BAUD")
            self.CanCat_send(b'\x01', CMD_CHANGE_BAUD_RESULT)

        elif cmd == CMD_PING:
            print(b'=CMD_PING=')
            self.log(b'=CMD_PING=')
            self.CanCat_send(data, CMD_PING_RESPONSE)

        elif cmd == CMD_CAN_MODE:
            print(b'=CMD_CAN_MODE:%r=' % data)
            self.log(b'=CMD_CAN_MODE:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_MODE_RESULT)

        elif cmd == CMD_CAN_BAUD:
            print(b'=CMD_CAN_BAUD:%r=' % data)
            self.log(b'=CMD_CAN_BAUD:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_BAUD_RESULT)

        elif cmd == CMD_CAN_SEND:
            print(b'=CMD_CAN_SEND:%r=' % data)
            self.log(b'=CMD_CAN_SEND:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_SEND_RESULT)

        elif cmd == CMD_SET_FILT_MASK:
            print(b'=CMD_SET_FILT_MASK:%r=' % data)
            self.log(b'=CMD_SET_FILT_MASK:%r=' % data)
            self.CanCat_send(b'\x01', CMD_SET_FILT_MASK_RESULT)

        elif cmd == CMD_CAN_SEND_ISOTP:
            print(b'=CMD_CAN_SEND_ISOTP:%r=' % data)
            self.log(b'=CMD_CAN_SEND_ISOTP:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_SEND_ISOTP_RESULT)

        elif cmd == CMD_CAN_RECV_ISOTP:
            print(b'=CMD_CAN_RECV_ISOTP:%r=' % data)
            self.log(b'=CMD_CAN_RECV_ISOTP:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_RECV_ISOTP_RESULT)

        elif cmd == CMD_CAN_SENDRECV_ISOTP:
            print(b'=CMD_CAN_SENDRECV_ISOTP:%r=' % data)
            self.log(b'=CMD_CAN_SENDRECV_ISOTP:%r=' % data)
            self.CanCat_send(b'\x01', CMD_CAN_SENDRECV_ISOTP_RESULT)

        elif cmd == CMD_PRINT_CAN_REGS:
            print(b'=CMD_PRINT_CAN_REGS:%r=' % data)
            self.log(b'=CMD_PRINT_CAN_REGS:%r=' % data)
            #self.CanCat_send(b'\x01', CMD_PRINT_CAN_REGS_RESULT)??

        else:
            print(b'===BAD COMMAND: %x : %r' % (cmd, data))
            self.log(b'===BAD COMMAND: %x : %r' % (cmd, data))
            self.CanCat_send(b'===BAD COMMAND: %x : %r' % (cmd, data), 3)



# Command line entry point for candump2cancat

import sys
import argparse
from cancat.utils import convert


def main():
    argv = sys.argv[1:]
                                                                                  
    parser = argparse.ArgumentParser(
            prog='candump2cancat',
            description='Utility to convert a candump log into a CanCat session')
    parser.add_argument('log', help='input candump log')
    parser.add_argument('output', help='output cancat session')
    args = parser.parse_args(argv)

    convert.candump2cancat(args.log, args.output)



from __future__ import print_function
from builtins import input

import sys
import time
import cancat
import struct
import threading

import cancat.iso_tp as cisotp

# In 11-bit CAN, an OBD2 tester typically sends requests with an ID of 7DF, and 
# can accept response messages on IDs 7E8 to 7EF, requests to a specific ECU can 
# be sent from ID 7E0 to 7E7.  So the non-OBD2 range normally ends at 7D7, 
# although I can't find a specific "standard" for this.
#
# In 29-bit CAN an OBD2 tester typically sends requests with an ID of 0x18DB33F1 
# where 0x18DBxxxx indicates this is an OBD2 message, 0x33 indicates this 
# message is for the OBD2 ECU(s), and 0xF1 is the tester.  Normal UDS messages 
# use a prefix of 0x18DAxxxx.
# 0xF1 is used as a tester address in normal UDS messaging as well.
ARBID_CONSTS = {
    '11bit': {
        'prefix': 0x700,
        'prefix_mask': 0xF00,
        'resp_offset': 8,  # rxid is normally the txid + 8
        'max_req_id': 0xF7,
        'obd2_broadcast': 0x7DF,
    },
    '29bit': {
        'prefix': 0x18DA0000,
        'prefix_mask': 0xFFFF0000,
        'destid_mask': 0x0000FF00,
        'destid_shift': 8,
        'srcid_mask': 0x000000FF,
        'tester': 0xF1,
        'obd2_broadcast': 0x18DA33F1,
    }
}

ISO_14229_DIDS = {
    0xF180: 'bootSoftwareIdentificationDataIdentifier',
    0xF181: 'applicationSoftwareIdentificationDataIdentifier',
    0xF182: 'applicationDataIdentificationDataIdentifier',
    0xF183: 'bootSoftwareFingerprintDataIdentifier',
    0xF184: 'applicationSoftwareFingerprintDataIdentifier',
    0xF185: 'applicationDataFingerprintDataIdentifier',
    0xF186: 'activeDiagnosticSessionDataIdentifier',
    0xF187: 'vehicleManufacturerSparePartNumberDataIdentifier',
    0xF188: 'vehicleManufacturerECUSoftwareNumberDataIdentifier',
    0xF189: 'vehicleManufacturerECUSoftwareVersionNumberDataIdentifier',
    0xF18A: 'systemSupplierIdentifierDataIdentifier',
    0xF18B: 'ECUManufacturingDateDataIdentifier',
    0xF18C: 'ECUSerialNumberDataIdentifier',
    0xF18D: 'supportedFunctionalUnitsDataIdentifier',
    0xF18E: 'vehicleManufacturerKitAssemblyPartNumberDataIdentifier',
    0xF190: 'VINDataIdentifier',
    0xF191: 'vehicleManufacturerECUHardwareNumberDataIdentifier',
    0xF192: 'systemSupplierECUHardwareNumberDataIdentifier',
    0xF193: 'systemSupplierECUHardwareVersionNumberDataIdentifier',
    0xF194: 'systemSupplierECUSoftwareNumberDataIdentifier',
    0xF195: 'systemSupplierECUSoftwareVersionNumberDataIdentifier',
    0xF196: 'exhaustRegulationOrTypeApprovalNumberDataIdentifier',
    0xF197: 'systemNameOrEngineTypeDataIdentifier',
    0xF198: 'repairShopCodeOrTesterSerialNumberDataIdentifier',
    0xF199: 'programmingDateDataIdentifier',
    0xF19A: 'calibrationRepairShopCodeOrCalibrationEquipmentSerialNumberData-',
    0xF19B: 'calibrationDateDataIdentifier',
    0xF19C: 'calibrationEquipmentSoftwareNumberDataIdentifier',
    0xF19D: 'ECUInstallationDateDataIdentifier',
    0xF19E: 'ODXFileDataIdentifier',
    0xF19F: 'entityDataIdentifier',
}

NEG_RESP_CODES = {
    0x10:'GeneralReject',
    0x11:'ServiceNotSupported',
    0x12:'SubFunctionNotSupported',
    0x13:'IncorrectMesageLengthOrInvalidFormat',
    0x14:'ResponseTooLong',
    0x21:'BusyRepeatRequest',
    0x22:'ConditionsNotCorrect',
    0x24:'RequestSequenceError',
    0x25:'NoResponseFromSubnetComponent',
    0x26:'FailurePreventsExecutionOfRequestedAction',
    0x31:'RequestOutOfRange',
    0x33:'SecurityAccessDenied',
    0x35:'InvalidKey',
    0x36:'ExceedNumberOfAttempts',
    0x37:'RequiredTimeDelayNotExpired',
    0x70:'UploadDownloadNotAccepted',
    0x71:'TransferDataSuspended',
    0x72:'GeneralProgrammingFailure',
    0x73:'WrongBlockSequenceCounter',
    0x78:'RequestCorrectlyReceived-ResponsePending',
    0x7e:'SubFunctionNotSupportedInActiveSession',
    0x7f:'ServiceNotSupportedInActiveSession',
    0x81:'RpmTooHigh',
    0x82:'RpmTooLow',
    0x83:'EngineIsRunning',
    0x84:'EngineIsNotRunning',
    0x85:'EngineRunTimeTooLow',
    0x86:'TemperatureTooHigh',
    0x87:'TemperatureTooLow',
    0x88:'VehicleSpeedTooHigh',
    0x89:'VehicleSpeedTooLow',
    0x8a:'ThrottlePedalTooHigh',
    0x8b:'ThrottlePedalTooLow',
    0x8c:'TransmissionRangeNotInNeutral',
    0x8d:'TransmissionRangeNotInGear',
    0x8f:'BrakeSwitchsNotClosed',
    0x90:'ShifterLeverNotInPark',
    0x91:'TorqueConverterClutchLocked',
    0x92:'VoltageTooHigh',
    0x93:'VoltageTooLow',
}

SVC_DIAGNOSTICS_SESSION_CONTROL =           0x10
SVC_ECU_RESET =                             0x11
SVC_CLEAR_DIAGNOSTICS_INFORMATION =         0x14
SVC_READ_DTC_INFORMATION =                  0x19
SVC_READ_DATA_BY_IDENTIFIER =               0x22
SVC_READ_MEMORY_BY_ADDRESS =                0x23
SVC_SECURITY_ACCESS =                       0x27
SVC_READ_DATA_BY_PERIODIC_IDENTIFIER =      0x2a
SVC_DYNAMICALLY_DEFINE_DATA_IDENTIFIER =    0x2c
SVC_WRITE_DATA_BY_IDENTIFIER =              0x2e
SVC_INPUT_OUTPUT_CONTROL_BY_IDENTIFIER =    0x2f
SVC_ROUTINE_CONTROL =                       0x31
SVC_REQUEST_DOWNLOAD =                      0x34
SVC_REQUEST_UPLOAD =                        0x35
SVC_TRANSFER_DATA =                         0x36
SVC_REQUEST_TRANSFER_EXIT =                 0x37
SVC_WRITE_MEMORY_BY_ADDRESS =               0x3d
SVC_TESTER_PRESENT =                        0x3e
SVC_NEGATIVE_RESPONSE =                     0x7f
SVC_CONTROL_DTC_SETTING =                   0x85

UDS_SVCS = { v:k for k,v in globals().items() if k.startswith('SVC_') }

POS_RESP_CODES = { (k|0x40) : "OK_" + v.lower() for k,v in UDS_SVCS.items() }
POS_RESP_CODES[0] = 'Success'

NEG_RESP_REPR = {}
for k,v in NEG_RESP_CODES.items():
    NEG_RESP_REPR[k] = 'ERR_' + v

RESP_CODES = {}
RESP_CODES.update(NEG_RESP_REPR)
RESP_CODES.update(POS_RESP_CODES)


class NegativeResponseException(Exception):
    def __init__(self, neg_code, svc, msg):
        self.neg_code = neg_code
        self.msg = msg
        self.svc = svc

    def __repr__(self):
        negresprepr = NEG_RESP_CODES.get(self.neg_code)
        return "NEGATIVE RESPONSE to 0x%x (%s):   ERROR 0x%x: %s   \tmsg: %s" % \
            (self.svc, UDS_SVCS.get(self.svc), self.neg_code, negresprepr, self.msg)

    def __str__(self):
        negresprepr = NEG_RESP_CODES.get(self.neg_code)
        return "NEGATIVE RESPONSE to 0x%x (%s):   ERROR 0x%x: %s   \tmsg: %s" % \
            (self.svc, UDS_SVCS.get(self.svc), self.neg_code, negresprepr, self.msg)


class UDS(object):
    def __init__(self, c, tx_arbid, rx_arbid=None, verbose=True, extflag=0, timeout=3.0):
        self.c = c
        self.t = None
        self.verbose = verbose
        self.extflag = extflag
        self.timeout = timeout

        if rx_arbid == None:
            rx_arbid = tx_arbid + 8 # by UDS spec

        self.tx_arbid = tx_arbid
        self.rx_arbid = rx_arbid

    def xmit_recv(self, data, extflag=0, count=1, service=None):
        msg, idx = self.c.ISOTPxmit_recv(self.tx_arbid, self.rx_arbid, data, extflag, self.timeout, count, service)

        # check if the response is something we know about and can help out
        if msg != None and len(msg):
            svc = data[0]
            svc_resp = msg[0]
            errcode = 0
            if len(msg) >= 3:
                errcode = msg[2]

            if svc_resp == svc + 0x40:
                if self.verbose:
                    print("Positive Response!")

            negresprepr = NEG_RESP_CODES.get(errcode)
            if negresprepr != None and svc_resp != svc + 0x40:
                if self.verbose > 1:
                    print(negresprepr + "\n")
                # TODO: Implement getting final message if ResponseCorrectlyReceivedResponsePending is received
                if errcode != 0x78: # Don't throw an exception for ResponseCorrectlyReceivedResponsePending
                    raise NegativeResponseException(errcode, svc, msg)
                else:
                    # Try again but increment the start index
                    msg, idx = self.c._isotp_get_msg(self.rx_arbid, start_index = idx+1, service = service, timeout = self.timeout)

        return msg

    def _do_Function(self, func, data=None, subfunc=None, service=None):
        if subfunc != None:
            omsg = struct.pack('>BB', func, subfunc)
        else:
            omsg = struct.pack('>B', func)

        if data != None:
            omsg += data

        msg = self.xmit_recv(omsg, extflag=self.extflag, service=service)
        return msg

    def SendTesterPresent(self):
        while self.TesterPresent is True:
            if self.TesterPresentRequestsResponse:
                self.c.CANxmit(self.tx_arbid, "023E000000000000".decode('hex'))
            else:
                self.c.CANxmit(self.tx_arbid, "023E800000000000".decode('hex'))
            time.sleep(2.0)

    def StartTesterPresent(self, request_response=True):
        self.TesterPresent = True
        self.TesterPresentRequestsResponse=request_response
        self.t = threading.Thread(target = self.SendTesterPresent)
        self.t.setDaemon(True)
        self.t.start()

    def StopTesterPresent(self):
        self.TesterPresent = False
        if self.t is not None:
            self.t.join(5.0)
            if self.t.isAlive():
                if self.verbose: 
                    print("Error killing Tester Present thread")
            self.t = None

    def DiagnosticSessionControl(self, session):
        currIdx = self.c.getCanMsgCount()
        return self._do_Function(SVC_DIAGNOSTICS_SESSION_CONTROL, chr(session), service=0x50)

    def ReadMemoryByAddress(self, address, size):
        currIdx = self.c.getCanMsgCount()
        return self._do_Function(SVC_READ_MEMORY_BY_ADDRESS, subfunc=0x24, data=struct.pack(">IH", address, size), service = 0x63)
        #return self.xmit_recv("\x23\x24" + struct.pack(">I", address) + struct.pack(">H", size), service = 0x63)

    def ReadDID(self, did):
        '''
        Read the Data Identifier specified from the ECU.

        Returns: The response ISO-TP message as a string
        '''
        msg = self._do_Function(SVC_READ_DATA_BY_IDENTIFIER, struct.pack('>H', did), service=0x62)
        #msg = self.xmit_recv("22".decode('hex') + struct.pack('>H', did), service=0x62)
        return msg

    def WriteDID(self, did, data):
        '''
        Write the Data Identifier specified from the ECU.

        Returns: The response ISO-TP message as a string
        '''
        msg = self._do_Function(SVC_WRITE_DATA_BY_IDENTIFIER,struct.pack('>H', did) + data, service=0x62)
        #msg = self.xmit_recv("22".decode('hex') + struct.pack('>H', did), service=0x62)
        return msg

    def RequestDownload(self, addr, data, data_format = 0x00, addr_format = 0x44):
        '''
        Assumes correct Diagnostics Session and SecurityAccess
        '''
        # Figure out the right address and data length formats
        pack_fmt_str = ">BB"
        try:
            pack_fmt_str += {1:"B", 2:"H", 4:"I"}.get(addr_format >> 4) + {1:"B", 2:"H", 4:"I"}.get(addr_format & 0xf)
        except TypeError:
            print("Cannot parse addressAndLengthFormatIdentifier", hex(addr_format))
            return None
        msg = self.xmit_recv("\x34" + struct.pack(pack_fmt_str, data_format, addr_format, addr, len(data)), extflag=self.extflag, service = 0x74)

        # Parse the response
        if msg[0] != 0x74:
            print("Error received: {}".format(msg.encode('hex')))
            return msg
        max_txfr_num_bytes = msg[1] >> 4 # number of bytes in the max tranfer length parameter
        max_txfr_len = 0
        for i in range(2,2+max_txfr_num_bytes):
            max_txfr_len <<= 8
            max_txfr_len += msg[i]

        # Transfer data
        data_idx = 0
        block_idx = 1
        while data_idx < len(data):
            msg = self.xmit_recv("\x36" + chr(block_idx) + data[data_idx:data_idx+max_txfr_len-2], extflag=self.extflag, service = 0x76)
            data_idx += max_txfr_len - 2
            block_idx += 1
            if block_idx > 0xff:
                block_idx = 0

            # error checking
            if msg is not None and msg[0] == 0x7f and msg[2] != 0x78:
                print("Error sending data: {}".format(msg.encode('hex')))
                return None
            if msg is None:
                print("Didn't get a response?")
                data_idx -= max_txfr_len - 2
                block_idx -= 1
                if block_idx == 0:
                    block_idx = 0xff

            # TODO: need to figure out how to get 2nd isotp message to verify that this worked

        # Send RequestTransferExit
        self._do_Function(SVC_REQUEST_TRANSFER_EXIT, service = 0x77)

    def readMemoryByAddress(self, address, length, lenlen=1, addrlen=4):
        '''
        Work in progress!
        '''
        if lenlen == 1:
            lfmt = "B"
        else:
            lfmt = "H"

        lenlenbyte = (lenlen << 4) | addrlen

        msg = self._do_Function(SVC_READ_MEMORY_BY_ADDRESS, data=struct.pack('<BI' + lfmt, lenlenbyte, address, length), service=0x63)

        return msg

    def writeMemoryByAddress(self, address, data, lenlen=1, addrlen=4):
        '''
        Work in progress!
        '''
        if lenlen == 1:
            lfmt = "B"
        else:
            lfmt = "H"

        lenlenbyte = (lenlen << 4) | addrlen

        data = struct.pack('<BI' + lfmt, lenlenbyte, address, length)
        #data = "3d".decode('hex') + struct.pack('<BI' + lfmt, lenlenbyte, address, length)

        msg = self._do_Function(SVC_WRITE_MEMORY_BY_ADDRESS, data=data, service=0x63)
        #msg = self.xmit_recv(data, service=0x63)

        return msg


    def RequestUpload(self, addr, length, data_format = 0x00, addr_format = 0x44):
        '''
        Work in progress!
        '''
        msg = self._do_Function(SVC_REQUEST_UPLOAD, subfunc=data_format, data = chr(addr_format) + struct.pack('>I', addr)[1:] +  struct.pack('>I', length)[1:])

        sid, lfmtid, maxnumblocks = struct.unpack('>BBH', msg[:4])

        output = []
        for loop in maxnumblocks:
            msg = self._do_Function(SVC_TRANSFER_DATA, subfunc=loop)
            output.append(msg)

            if len(msg) and msg[0] != '\x76':
                print("FAILURE TO DOWNLOAD ALL.  Returning what we have so far (including error message)")
                return output

        msg = self._do_Function(SVC_REQUEST_TRANSFER_EXIT)
        if len(msg) and msg[0] != '\x77':
            print("FAILURE TO EXIT CLEANLY.  Returning what we received.")

        return output

    def EcuReset(self, rst_type=0x1):
        return self._do_Function(SVC_ECU_RESET, subfunc=rst_type)

    def ClearDiagnosticInformation(self):
        pass
    def ReadDTCInfomation(self):
        pass
    def ReadDataByPeriodicIdentifier(self, pdid):
        pass
    def DynamicallyDefineDataIdentifier(self):
        pass
    def InputOutputControlByIdentifier(self, iodid):
        pass
    def RoutineControl(self, rid):
        pass
    def TransferData(self, did):
        pass
    def RequestTransferExit(self):
        pass
    def ControlDTCSetting(self):
        pass



    def ScanDIDs(self, start=0, end=0x10000, delay=0):
        success = []
        try:
            for x in range(start, end):
                try:
                    if self.verbose:
                        sys.stderr.write(' %x ' % x)

                    val = self.ReadDID(x)
                    success.append((x, val))

                except KeyboardInterrupt:
                    raise

                except Exception as e:
                    if self.verbose > 1:
                        print(e)

                time.sleep(delay)

        except KeyboardInterrupt:
            print("Stopping Scan during DID 0x%x " % x)
            return success

        return success


    def SecurityAccess(self, level, key = ""):
        """Send and receive the UDS messages to switch SecurityAccess levels.
            @level = the SecurityAccess level to switch to
            @key = a SecurityAccess algorithm specific key
        """
        msg = self._do_Function(SVC_SECURITY_ACCESS, subfunc=level, service = 0x67)
        if msg is None:
            return msg
        if msg[0] == 0x7f:
            print("Error getting seed:", msg.encode('hex'))

        else:
            seed = msg[2:]
            hexified_seed = " ".join(x.encode('hex') for x in seed)
            key = str(bytearray(self._key_from_seed(hexified_seed, key)))

            msg = self._do_Function(SVC_SECURITY_ACCESS, subfunc=level+1, data=key, service = 0x67)
            return msg


    def _key_from_seed(self, seed, secret):
        """Generates the key for a specific SecurityAccess seed request.
            @seed = the SecurityAccess seed received from the ECU.  Formatted
                    as a hex string with spaces between each seed byte.
            @secret = a SecurityAccess algorithm specific key
           Returns the key, as a string of key bytes.
        """
        print("Not implemented in this class")
        return 0


def printUDSSession(c, tx_arbid, rx_arbid=None, paginate=45):
    if rx_arbid == None:
        rx_arbid = tx_arbid + 8 # by UDS spec

    msgs = [msg for msg in c.genCanMsgs(arbids=[tx_arbid, rx_arbid])]

    msgs_idx = 0

    linect = 1
    while msgs_idx < len(msgs):
        arbid, isotpmsg, count = cisotp.msg_decode(msgs, msgs_idx)
        #print("Message: (%s:%s) \t %s" % (count, msgs_idx, isotpmsg.encode('hex')))
        svc = isotpmsg[0]
        mtype = (RESP_CODES, UDS_SVCS)[arbid==tx_arbid].get(svc, '')

        print("Message: (%s:%s) \t %-30s %s" % (count, msgs_idx, isotpmsg.encode('hex'), mtype))
        msgs_idx += count

        if paginate:
            if (linect % paginate)==0:
                input("%x)  PRESS ENTER" % linect)

        linect += 1
